!===============================================================================
!
! rhmc_poly.F90 - polynomial filtering with RHMC
!
!-------------------------------------------------------------------------------
!
! Subroutines for the use of polynomial filtering
! when simulating singleton flavours of quarks (e.g. strange).
!
! Author: Taylor Haar
! Last modified: 21/03/2017
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine cheby_sqrt(porder, c, z_k)
  ! Generates a polynomial approximation of order `porder` to 1/sqrt(z)
  use module_input
  implicit none

  integer, intent(in)                       :: porder
  REAL, intent(out)                         :: c
  COMPLEX, dimension(porder), intent(inout) :: z_k

  ! Uses a pre-computed Chebyshev approximation
  ! Might calculate these directly in future versions

  if (porder == 4) then
     ! ! Generated Chebyshev polynomial for n=4, domain=[1e-4, 3]
     c = 0.3690967
     z_k(1) = cmplx(0.87173547, -0.74047338)
     z_k(2) = cmplx(2.91910349, -0.57611755)
     z_k(3) = conjg(z_k(2))
     z_k(4) = conjg(z_k(1))
  elseif (porder == 8) then
     ! Generated Chebyshev polynomial for n=8, domain=[1e-4, 3]
     c = 0.64084772
     z_k(1) = cmplx(0.26360806, -0.26645310)
     z_k(2) = cmplx(1.18759024, -0.53056798)
     z_k(3) = cmplx(2.28944057, -0.51680453)
     z_k(4) = cmplx(3.02135855, -0.21265364)
     z_k(5) = conjg(z_k(4))
     z_k(6) = conjg(z_k(3))
     z_k(7) = conjg(z_k(2))
     z_k(8) = conjg(z_k(1))
  elseif (porder == 10) then
     ! Generated Chebyshev polynomial for n=10, domain=[1e-4, 3]
     c = 0.92919461
     z_k(1) = cmplx(0.17565703, -0.18226592)
     z_k(2) = cmplx(0.82192124, -0.39959941)
     z_k(3) = cmplx(1.70518054, -0.48578443)
     z_k(4) = cmplx(2.53050228, -0.39235476)
     z_k(5) = cmplx(3.02482769, -0.14990778)
     z_k(6) = conjg(z_k(5))
     z_k(7) = conjg(z_k(4))
     z_k(8) = conjg(z_k(3))
     z_k(9) = conjg(z_k(2))
     z_k(10) = conjg(z_k(1))
  elseif (porder == 12) then
     ! Generated Chebyshev polynomial for n=12, domain=[1e-4, 3]
     c = 1.39354620
     z_k(1) = cmplx(0.12539113, -0.13217127)
     z_k(2) = cmplx(0.59932457, -0.30535433)
     z_k(3) = cmplx(1.29428764, -0.41327116)
     z_k(4) = cmplx(2.04341241, -0.41480569)
     z_k(5) = cmplx(2.66931913, -0.30487512)
     z_k(6) = cmplx(3.02416045, -0.11183642)
     z_k(7) = conjg(z_k(6))
     z_k(8) = conjg(z_k(5))
     z_k(9) = conjg(z_k(4))
     z_k(10) = conjg(z_k(3))
     z_k(11) = conjg(z_k(2))
     z_k(12) = conjg(z_k(1))
  elseif (porder == 16) then
     ! Generated Chebyshev polynomial for n=16, domain=[1e-4, 3]
     c = 3.34612239
     z_k(1) = cmplx(0.07304946, -0.07842398)
     z_k(2) = cmplx(0.35695267, -0.19094706)
     z_k(3) = cmplx(0.80273188, -0.28580925)
     z_k(4) = cmplx(1.34700286, -0.34107299)
     z_k(5) = cmplx(1.91432159, -0.34475571)
     z_k(6) = cmplx(2.42644618, -0.29463669)
     z_k(7) = cmplx(2.81284470, -0.19790804)
     z_k(8) = cmplx(3.02032343, -0.06967568)
     z_k(9) = conjg(z_k(8))
     z_k(10) = conjg(z_k(7))
     z_k(11) = conjg(z_k(6))
     z_k(12) = conjg(z_k(5))
     z_k(13) = conjg(z_k(4))
     z_k(14) = conjg(z_k(3))
     z_k(15) = conjg(z_k(2))
     z_k(16) = conjg(z_k(1))
  else
     call die("Implementation pending:" // &
          " please choose nf1_porder=[4|8|10|12|16].")
  endif

end subroutine cheby_sqrt

!===============================================================================
     


  
  
