
/*******************************************************************************
*
* File dh.h
*
* Copyright (C) 2007 Bjoern Leder & Rainer Sommer
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Type definitions and macros for fast manipulation  of
* SU(3) matrices, SU(3) vectors and Dirac spinors exploiting the Double Hummer
* unit of BlueGene/L
*
*******************************************************************************/

#ifndef DH_H
#define DH_H

/*******************************************************************************
*
* SU(3) vector manipulations
*
*******************************************************************************/


/*** Declaration of double _Complex variables ***/

/*
	Temporary variables to store 1*3 spin and color components of 4 spinors.
	In a function f(r,s) of two spinors r and s the variables dcra and dcs
	will store the spinor components of r and s. The variables ua and ta
	will be used to store intermediary results.
*/
#define _dh_vector_declare2(a) \
                double _Complex  dcr##a##1,dcr##a##2,dcr##a##3;\
                double _Complex  dcs##a##1,dcs##a##2,dcs##a##3;\
                double _Complex  t##a##1,t##a##2,t##a##3;\
                double _Complex  u##a##1,u##a##2,u##a##3

/*
	Temporary variables to store 1*3 spin and color components of 1 spinor.
	No extra variables for intermediary results.
*/
#define _dh_vector_declare1(a) \
                double _Complex  dcr##a##1,dcr##a##2,dcr##a##3

/*
	Temporary variables to store 1*3 spin and color components of 2 spinors.
*/
#define _dh_vector_declareu(a) \
                double _Complex  dcr##a##1,dcr##a##2,dcr##a##3;\
                double _Complex  u##a##1,u##a##2,u##a##3

/*
	Temporary variables to store 1*3 spin and color components of 3 spinors.
*/
#define _dh_vector_declare(a) \
                double _Complex  dcr##a##1,dcr##a##2,dcr##a##3;\
                double _Complex  t##a##1,t##a##2,t##a##3;\
                double _Complex  u##a##1,u##a##2,u##a##3



/*** Loading in parallel all color components ***

	Most of these macros come in two flavors: one for double precision (_dh_...)
	and one for single precision (_dhs_...) numbers.

***/

/*
	Load the three color components of r and s into
	temporary variables dcra1...dcra3 and dcsa1...dcsa3.
*/
#define _dh_vector_load2(r,s,a) \
                dcr##a##1 = __lfpd(&(r).c1.re); \
                dcr##a##2 = __lfpd(&(r).c2.re); \
                dcr##a##3 = __lfpd(&(r).c3.re); \
                dcs##a##1 = __lfpd(&(s).c1.re); \
                dcs##a##2 = __lfpd(&(s).c2.re); \
                dcs##a##3 = __lfpd(&(s).c3.re)

#define _dhs_vector_load2(r,s,a) \
                dcr##a##1 = __lfps(&(r).c1.re); \
                dcr##a##2 = __lfps(&(r).c2.re); \
                dcr##a##3 = __lfps(&(r).c3.re); \
                dcs##a##1 = __lfps(&(s).c1.re); \
                dcs##a##2 = __lfps(&(s).c2.re); \
                dcs##a##3 = __lfps(&(s).c3.re)

/*
	Load the three color components of r into
	temporary variables dcra1...dcra3.
*/
#define _dh_vector_load(r,a) \
                dcr##a##1 = __lfpd(&(r).c1.re); \
                dcr##a##2 = __lfpd(&(r).c2.re); \
                dcr##a##3 = __lfpd(&(r).c3.re)

#define _dhs_vector_load(r,a) \
                dcr##a##1 = __lfps(&(r).c1.re); \
                dcr##a##2 = __lfps(&(r).c2.re); \
                dcr##a##3 = __lfps(&(r).c3.re)

/*
	Load, alternating, the three color components of r and s into
	temporary variables dcra1...dcra3 and dcsa1...dcsa3.
*/
#define _dh_vector_pload(r,a,s,b) \
                dcr##a##1 = __lfpd(&(r).c1.re); \
                dcr##b##1 = __lfpd(&(s).c1.re); \
                dcr##a##2 = __lfpd(&(r).c2.re); \
                dcr##b##2 = __lfpd(&(s).c2.re); \
                dcr##a##3 = __lfpd(&(r).c3.re); \
                dcr##b##3 = __lfpd(&(s).c3.re)

/*
	Load the three color components of s into
	temporary variables dcsa1...dcsa3.
*/
#define _dh_vector_loads(s,a) \
                dcs##a##1 = __lfpd(&(s).c1.re); \
                dcs##a##2 = __lfpd(&(s).c2.re); \
                dcs##a##3 = __lfpd(&(s).c3.re)

/*
	Load the three color components of s into
	temporary variables ua1...ua3.
*/
#define _dh_vector_loadu(s,a) \
                u##a##1 = __lfpd(&(s).c1.re); \
                u##a##2 = __lfpd(&(s).c2.re); \
                u##a##3 = __lfpd(&(s).c3.re)
#define _dhs_vector_loadu(s,a) \
                u##a##1 = __lfps(&(s).c1.re); \
                u##a##2 = __lfps(&(s).c2.re); \
                u##a##3 = __lfps(&(s).c3.re)


/*** Storing in parallel all color components ***

	Most of these macros come in two flavors: one for double precision (_dh_...)
	and one for single precision (_dhs_...) numbers.

***/

/*
	Store the three temporary variables dcra1...dcra3 into
	the color components of r.
*/
#define _dh_vector_store(r,a) \
                __stfpd(&(r).c1.re,  dcr##a##1); \
                __stfpd(&(r).c2.re,  dcr##a##2); \
                __stfpd(&(r).c3.re,  dcr##a##3)

#define _dhs_vector_store(r,a) \
                __stfps(&(r).c1.re,  dcr##a##1); \
                __stfps(&(r).c2.re,  dcr##a##2); \
                __stfps(&(r).c3.re,  dcr##a##3)

/*
	Store the three temporary variables ua1...ua3 into
	the color components of r.
*/
#define _dh_vector_storeu(r,a) \
                __stfpd(&(r).c1.re,  u##a##1); \
                __stfpd(&(r).c2.re,  u##a##2); \
                __stfpd(&(r).c3.re,  u##a##3)

#define _dhs_vector_storeu(r,a) \
                __stfps(&(r).c1.re,  u##a##1); \
                __stfps(&(r).c2.re,  u##a##2); \
                __stfps(&(r).c3.re,  u##a##3)

/*
	Store the three temporary variables ta1...ta3 into
	the color components of r.
*/
#define _dh_vector_storet(r,a) \
                __stfpd(&(r).c1.re,  t##a##1); \
                __stfpd(&(r).c2.re,  t##a##2); \
                __stfpd(&(r).c3.re,  t##a##3)

#define _dhs_vector_storet(r,a) \
                __stfps(&(r).c1.re,  t##a##1); \
                __stfps(&(r).c2.re,  t##a##2); \
                __stfps(&(r).c3.re,  t##a##3)


/*** Arithmetic operations ***/

/*
	ua = - dcra
*/
#define _dh_vector_neg(a) \
                u##a##1 = __fpneg(dcr##a##1); \
                u##a##2 = __fpneg(dcr##a##2); \
                u##a##3 = __fpneg(dcr##a##3)

/*	
	ua1 = c*dcra1  (c real)
	ua2 = c*dcra2
	ua3 = c*dcra3
	ta1 = c*dcsa1  (c real)
	ta2 = c*dcsa2
	ta3 = c*dcsa3
*/
#define _dh_vector_mul(c,a) \
                u##a##1 = __fxpmul(dcr##a##1, __creal(c)); \
                u##a##2 = __fxpmul(dcr##a##2, __creal(c)); \
                u##a##3 = __fxpmul(dcr##a##3, __creal(c)); \
                t##a##1 = __fxpmul(dcs##a##1, __creal(c)); \
                t##a##2 = __fxpmul(dcs##a##2, __creal(c)); \
                t##a##3 = __fxpmul(dcs##a##3, __creal(c))

/*
	ua1 = c*dcra1  (c complex)
	ua2 = c*dcra2
	ua3 = c*dcra3
*/
#define _dh_vector_mulc2(c,a,b) \
                t##a##1 = __fxpmul(dcr##a##1, __creal(c)); \
                t##a##2 = __fxpmul(dcr##a##2, __creal(c)); \
                t##a##3 = __fxpmul(dcr##a##3, __creal(c)); \
                u##a##1 = __fxcxnpma(  t##a##1, dcr##a##1, __cimag(c)); \
                u##a##2 = __fxcxnpma(  t##a##2, dcr##a##2, __cimag(c)); \
                u##a##3 = __fxcxnpma(  t##a##3, dcr##a##3, __cimag(c)); \
					 \
                t##b##1 = __fxpmul(dcr##b##1, __creal(c)); \
                t##b##2 = __fxpmul(dcr##b##2, __creal(c)); \
                t##b##3 = __fxpmul(dcr##b##3, __creal(c)); \
                u##b##1 = __fxcxnpma(  t##b##1, dcr##b##1, __cimag(c)); \
                u##b##2 = __fxcxnpma(  t##b##2, dcr##b##2, __cimag(c)); \
                u##b##3 = __fxcxnpma(  t##b##3, dcr##b##3, __cimag(c));

/*
	r.c1 = dcra1 + z*dcsa1 (z complex)
	r.c2 = dcra2 + z*dcsa2
	r.c3 = dcra3 + z*dcsa3
*/
#define _dh_vector_mulc_assign(r,z,a) \
                t##a##1 = __fxcpmadd(dcr##a##1, dcs##a##1, __creal(z)); \
                t##a##2 = __fxcpmadd(dcr##a##2, dcs##a##2, __creal(z)); \
                t##a##3 = __fxcpmadd(dcr##a##3, dcs##a##3, __creal(z)); \
                u##a##1 = __fxcxnpma(  t##a##1, dcs##a##1, __cimag(z)); \
                u##a##2 = __fxcxnpma(  t##a##2, dcs##a##2, __cimag(z)); \
                u##a##3 = __fxcxnpma(  t##a##3, dcs##a##3, __cimag(z)); \
                __stfpd(&(r).c1.re,  u##a##1); \
                __stfpd(&(r).c2.re,  u##a##2); \
                __stfpd(&(r).c3.re,  u##a##3)

/*
	ua1 = dcra1 + z*dcsa1 (z complex)
	ua2 = dcra2 + z*dcsa2
	ua3 = dcra3 + z*dcsa3 and equiv. for ub ...
*/
#define _dh_vector_mulc_assign2(z,a,b) \
                t##a##1 = __fxcpmadd(dcr##a##1, dcs##a##1, __creal(z)); \
                t##a##2 = __fxcpmadd(dcr##a##2, dcs##a##2, __creal(z)); \
                t##a##3 = __fxcpmadd(dcr##a##3, dcs##a##3, __creal(z)); \
					 \
                t##b##1 = __fxcpmadd(dcr##b##1, dcs##b##1, __creal(z)); \
                t##b##2 = __fxcpmadd(dcr##b##2, dcs##b##2, __creal(z)); \
                t##b##3 = __fxcpmadd(dcr##b##3, dcs##b##3, __creal(z)); \
					 \
                u##a##1 = __fxcxnpma(  t##a##1, dcs##a##1, __cimag(z)); \
                u##a##2 = __fxcxnpma(  t##a##2, dcs##a##2, __cimag(z)); \
                u##a##3 = __fxcxnpma(  t##a##3, dcs##a##3, __cimag(z)); \
					 \
                u##b##1 = __fxcxnpma(  t##b##1, dcs##b##1, __cimag(z)); \
                u##b##2 = __fxcxnpma(  t##b##2, dcs##b##2, __cimag(z)); \
                u##b##3 = __fxcxnpma(  t##b##3, dcs##b##3, __cimag(z))

/*
	ua1 += z*dcra1
	ua2 += z*dcra2
	ua3 += z*dcra3 and equiv. for ub ...
*/
#define _dh_vector_mulc_add2(z,a,b) \
                t##a##1 = __fxcpmadd(u##a##1, dcr##a##1, __creal(z)); \
                t##a##2 = __fxcpmadd(u##a##2, dcr##a##2, __creal(z)); \
                t##a##3 = __fxcpmadd(u##a##3, dcr##a##3, __creal(z)); \
                t##b##1 = __fxcpmadd(u##b##1, dcr##b##1, __creal(z)); \
                t##b##2 = __fxcpmadd(u##b##2, dcr##b##2, __creal(z)); \
                t##b##3 = __fxcpmadd(u##b##3, dcr##b##3, __creal(z)); \
					 \
                u##a##1 = __fxcxnpma(  t##a##1, dcr##a##1, __cimag(z)); \
                u##a##2 = __fxcxnpma(  t##a##2, dcr##a##2, __cimag(z)); \
                u##a##3 = __fxcxnpma(  t##a##3, dcr##a##3, __cimag(z)); \
                u##b##1 = __fxcxnpma(  t##b##1, dcr##b##1, __cimag(z)); \
                u##b##2 = __fxcxnpma(  t##b##2, dcr##b##2, __cimag(z)); \
                u##b##3 = __fxcxnpma(  t##b##3, dcr##b##3, __cimag(z))

/*
	Scalar product ua1+ua2+ua3=(ra,sa)
*/
#define _dh_vector_prod(a) \
                t##a##1 = __fxpmul(dcs##a##1, __creal(dcr##a##1)); \
                t##a##2 = __fxpmul(dcs##a##2, __creal(dcr##a##2)); \
                t##a##3 = __fxpmul(dcs##a##3, __creal(dcr##a##3)); \
                u##a##1 = __fxcxnsma(  t##a##1, dcs##a##1, __cimag(dcr##a##1)); \
                u##a##2 = __fxcxnsma(  t##a##2, dcs##a##2, __cimag(dcr##a##2)); \
                u##a##3 = __fxcxnsma(  t##a##3, dcs##a##3, __cimag(dcr##a##3))

/*
	Scalar product ua1+ua2+ua3=(ra,sa), ub1+ub2+ub3=(rb,sb)
*/
#define _dh_vector_prod2(a,b) \
                t##a##1 = __fxpmul(dcs##a##1, __creal(dcr##a##1)); \
                t##a##2 = __fxpmul(dcs##a##2, __creal(dcr##a##2)); \
                t##a##3 = __fxpmul(dcs##a##3, __creal(dcr##a##3)); \
                t##b##1 = __fxpmul(dcs##b##1, __creal(dcr##b##1)); \
                t##b##2 = __fxpmul(dcs##b##2, __creal(dcr##b##2)); \
                t##b##3 = __fxpmul(dcs##b##3, __creal(dcr##b##3)); \
					 \
                u##a##1 = __fxcxnsma(  t##a##1, dcs##a##1, __cimag(dcr##a##1)); \
                u##a##2 = __fxcxnsma(  t##a##2, dcs##a##2, __cimag(dcr##a##2)); \
                u##a##3 = __fxcxnsma(  t##a##3, dcs##a##3, __cimag(dcr##a##3)); \
                u##b##1 = __fxcxnsma(  t##b##1, dcs##b##1, __cimag(dcr##b##1)); \
                u##b##2 = __fxcxnsma(  t##b##2, dcs##b##2, __cimag(dcr##b##2)); \
                u##b##3 = __fxcxnsma(  t##b##3, dcs##b##3, __cimag(dcr##b##3))

/*
	Vector squared ua1+ua2+ua3=(ra,ra), ub1+ub2+ub3=(rb,rb)
*/
#define _dh_vector_square2(a,b) \
                t##a##1 = __fxpmul(dcr##a##1, __creal(dcr##a##1)); \
                t##a##2 = __fxpmul(dcr##a##2, __creal(dcr##a##2)); \
                t##a##3 = __fxpmul(dcr##a##3, __creal(dcr##a##3)); \
                t##b##1 = __fxpmul(dcr##b##1, __creal(dcr##b##1)); \
                t##b##2 = __fxpmul(dcr##b##2, __creal(dcr##b##2)); \
                t##b##3 = __fxpmul(dcr##b##3, __creal(dcr##b##3)); \
					 \
                u##a##1 = __fxcxnsma(  t##a##1, dcr##a##1, __cimag(dcr##a##1)); \
                u##a##2 = __fxcxnsma(  t##a##2, dcr##a##2, __cimag(dcr##a##2)); \
                u##a##3 = __fxcxnsma(  t##a##3, dcr##a##3, __cimag(dcr##a##3)); \
                u##b##1 = __fxcxnsma(  t##b##1, dcr##b##1, __cimag(dcr##b##1)); \
                u##b##2 = __fxcxnsma(  t##b##2, dcr##b##2, __cimag(dcr##b##2)); \
                u##b##3 = __fxcxnsma(  t##b##3, dcr##b##3, __cimag(dcr##b##3))

/*
	z=ua1+ua2+ua3
*/
#define _dh_vector_addu(z,a) \
                t##a##1 = __fpadd(u##a##1,u##a##2); \
		t##a##2 = __fpadd(t##a##1,u##a##3); \
		z = __fpadd(z,t##a##2)

/*
	z=(dcra,dcsa)+(dcrb,dcsb)
*/
#define _dh_vector_prod_add2(z,a,b) \
                t##a##1 = __fxpmul(dcs##a##1, __creal(dcr##a##1)); \
                t##b##1 = __fxpmul(dcs##b##1, __creal(dcr##b##1)); \
                t##a##2 = __fxcxnsma(  t##a##1, dcs##a##1, __cimag(dcr##a##1)); \
                t##b##2 = __fxcxnsma(  t##b##1, dcs##b##1, __cimag(dcr##b##1)); \
                t##a##3 = __fxcpmadd(  t##a##2, dcs##a##2, __creal(dcr##a##2)); \
                t##b##3 = __fxcpmadd(  t##b##2, dcs##b##2, __creal(dcr##b##2)); \
                u##a##1 = __fxcxnsma(  t##a##3, dcs##a##2, __cimag(dcr##a##2)); \
                u##b##1 = __fxcxnsma(  t##b##3, dcs##b##2, __cimag(dcr##b##2)); \
                u##a##2 = __fxcpmadd(  u##a##1, dcs##a##3, __creal(dcr##a##3)); \
                u##b##2 = __fxcpmadd(  u##b##1, dcs##b##3, __creal(dcr##b##3)); \
                u##a##3 = __fxcxnsma(  u##a##2, dcs##a##3, __cimag(dcr##a##3)); \
                u##b##3 = __fxcxnsma(  u##b##2, dcs##b##3, __cimag(dcr##b##3)); \
		z = __fpadd(u##a##3,u##b##3)

/*
	z=(dcra,dcsa)+(dcrb,dcsb)+(dcrc,dcsc)+(dcrd,dcsd)
*/
#define _dh_vector_prod_add4(z,a,b,c,d) \
                t##a##1 = __fxpmul(dcs##a##1, __creal(dcr##a##1)); \
                t##b##1 = __fxpmul(dcs##b##1, __creal(dcr##b##1)); \
                t##c##1 = __fxpmul(dcs##c##1, __creal(dcr##c##1)); \
                t##d##1 = __fxpmul(dcs##d##1, __creal(dcr##d##1)); \
					 \
                t##a##2 = __fxcxnsma(  t##a##1, dcs##a##1, __cimag(dcr##a##1)); \
                t##b##2 = __fxcxnsma(  t##b##1, dcs##b##1, __cimag(dcr##b##1)); \
                t##c##2 = __fxcxnsma(  t##c##1, dcs##c##1, __cimag(dcr##c##1)); \
                t##d##2 = __fxcxnsma(  t##d##1, dcs##d##1, __cimag(dcr##d##1)); \
					 \
                t##a##3 = __fxcpmadd(  t##a##2, dcs##a##2, __creal(dcr##a##2)); \
                t##b##3 = __fxcpmadd(  t##b##2, dcs##b##2, __creal(dcr##b##2)); \
                t##c##3 = __fxcpmadd(  t##c##2, dcs##c##2, __creal(dcr##c##2)); \
                t##d##3 = __fxcpmadd(  t##d##2, dcs##d##2, __creal(dcr##d##2)); \
					 \
                u##a##1 = __fxcxnsma(  t##a##3, dcs##a##2, __cimag(dcr##a##2)); \
                u##b##1 = __fxcxnsma(  t##b##3, dcs##b##2, __cimag(dcr##b##2)); \
                u##c##1 = __fxcxnsma(  t##c##3, dcs##c##2, __cimag(dcr##c##2)); \
                u##d##1 = __fxcxnsma(  t##d##3, dcs##d##2, __cimag(dcr##d##2)); \
					 \
                u##a##2 = __fxcpmadd(  u##a##1, dcs##a##3, __creal(dcr##a##3)); \
                u##b##2 = __fxcpmadd(  u##b##1, dcs##b##3, __creal(dcr##b##3)); \
                u##c##2 = __fxcpmadd(  u##c##1, dcs##c##3, __creal(dcr##c##3)); \
                u##d##2 = __fxcpmadd(  u##d##1, dcs##d##3, __creal(dcr##d##3)); \
					 \
                u##a##3 = __fxcxnsma(  u##a##2, dcs##a##3, __cimag(dcr##a##3)); \
                u##b##3 = __fxcxnsma(  u##b##2, dcs##b##3, __cimag(dcr##b##3)); \
                u##c##3 = __fxcxnsma(  u##c##2, dcs##c##3, __cimag(dcr##c##3)); \
                u##d##3 = __fxcxnsma(  u##d##2, dcs##d##3, __cimag(dcr##d##3)); \
					 \
		t##a##1 = __fpadd(u##a##3,u##b##3); \
		t##a##2 = __fpadd(u##c##3,u##d##3); \
		z = __fpadd(t##a##1,t##a##2)

/*
	uc1=dcra1+dcrb1
	uc2=dcra2+dcrb2
	uc3=dcra3+dcrb3
*/
#define _dh_vector_add(c,a,b) \
		u##c##1 = __fpadd(dcr##a##1,dcr##b##1); \
		u##c##2 = __fpadd(dcr##a##2,dcr##b##2); \
		u##c##3 = __fpadd(dcr##a##3,dcr##b##3)

/*
	uc1=dcra1-dcrb1
	uc2=dcra2-dcrb2
	uc3=dcra3-dcrb3
*/
#define _dh_vector_sub(c,a,b) \
		u##c##1 = __fpsub(dcr##a##1,dcr##b##1); \
		u##c##2 = __fpsub(dcr##a##2,dcr##b##2); \
		u##c##3 = __fpsub(dcr##a##3,dcr##b##3)

/*
	uc1=dcra1+i*dcrb1
	uc2=dcra2+i*dcrb2
	uc3=dcra3+i*dcrb3
*/
#define _dh_vector_i_add(c,a,b) \
                u##c##1 = __fxcxnpma(dcr##a##1,dcr##b##1,1.0); \
		u##c##2 = __fxcxnpma(dcr##a##2,dcr##b##2,1.0); \
		u##c##3 = __fxcxnpma(dcr##a##3,dcr##b##3,1.0)

/*
	uc1=dcra1-i*dcrb1
	uc2=dcra2-i*dcrb2
	uc3=dcra3-i*dcrb3
*/
#define _dh_vector_i_sub(c,a,b) \
                u##c##1 = __fxcxnsma(dcr##a##1,dcr##b##1,1.0); \
		u##c##2 = __fxcxnsma(dcr##a##2,dcr##b##2,1.0); \
		u##c##3 = __fxcxnsma(dcr##a##3,dcr##b##3,1.0)

/*
	dcra1=dcrb1
	dcra2=dcrb2
	dcra3=dcrb3
*/
#define _dh_vector_assign(a,b) \
		dcr##a##3 = dcr##b##3; \
                dcr##a##1 = dcr##b##1; \
		dcr##a##2 = dcr##b##2

/*
	dcra1+=dcrb1
	dcra2+=dcrb2
	dcra3+=dcrb3
*/
#define _dh_vector_add_assign(a,b) \
 		dcr##a##3 = __fpadd(dcr##a##3,dcr##b##3); \
		dcr##a##1 = __fpadd(dcr##a##1,dcr##b##1); \
		dcr##a##2 = __fpadd(dcr##a##2,dcr##b##2)

/*
	dcra1=dcrb1+uc1
	dcra2=dcrb2+uc2
	dcra3=dcrb3+uc3
*/
#define _dh_vector_add_assign2(a,b,c) \
 		dcr##a##1 = __fpadd(dcr##b##1,u##c##1); \
		dcr##a##2 = __fpadd(dcr##b##2,u##c##2); \
		dcr##a##3 = __fpadd(dcr##b##3,u##c##3)

/*
	dcra1-=dcrb1
	dcra2-=dcrb2
	dcra3-=dcrb3
*/
#define _dh_vector_sub_assign(a,b) \
 		dcr##a##3 = __fpsub(dcr##a##3,dcr##b##3); \
		dcr##a##1 = __fpsub(dcr##a##1,dcr##b##1); \
		dcr##a##2 = __fpsub(dcr##a##2,dcr##b##2)

/*
	dcra1+=i*dcrb1
	dcra2+=i*dcrb2
	dcra3+=i*dcrb3
*/
#define _dh_vector_i_add_assign(a,b) \
 		dcr##a##3 = __fxcxnpma(dcr##a##3,dcr##b##3,1.0); \
		dcr##a##1 = __fxcxnpma(dcr##a##1,dcr##b##1,1.0); \
		dcr##a##2 = __fxcxnpma(dcr##a##2,dcr##b##2,1.0)

/*
	dcra1-=i*dcrb1
	dcra2-=i*dcrb2
	dcra3-=i*dcrb3
*/
#define _dh_vector_i_sub_assign(a,b) \
 		dcr##a##3 = __fxcxnsma(dcr##a##3,dcr##b##3,1.0); \
		dcr##a##1 = __fxcxnsma(dcr##a##1,dcr##b##1,1.0); \
		dcr##a##2 = __fxcxnsma(dcr##a##2,dcr##b##2,1.0)

/*
	dcra1*=c
	dcra2*=c
	dcra3*=c
*/
#define _dh_vector_mul_assign(a,c) \
					dcr##a##3 = __fxpmul(dcr##a##3, c); \
					dcr##a##1 = __fxpmul(dcr##a##1, c); \
					dcr##a##2 = __fxpmul(dcr##a##2, c)



/*******************************************************************************
*
* SU(3) matrix manipulations
*
*******************************************************************************/

/*
	Temporary variables to store su(3) matrix
*/
#define _dh_su3_declare(a) \
                double _Complex  su##a##11,su##a##12,su##a##13,\
				 su##a##21,su##a##22,su##a##23,\
				 su##a##31,su##a##32,su##a##33

/*
	Load su(3) matrix into temporary variables
*/
#define _dh_su3_load(u,a) \
                su##a##11 = __lfpd(&(u).c11.re);\
                su##a##22 = __lfpd(&(u).c22.re);\
                su##a##33 = __lfpd(&(u).c33.re);\
                su##a##12 = __lfpd(&(u).c12.re);\
                su##a##23 = __lfpd(&(u).c23.re);\
                su##a##31 = __lfpd(&(u).c31.re);\
                su##a##21 = __lfpd(&(u).c21.re);\
                su##a##32 = __lfpd(&(u).c32.re);\
                su##a##13 = __lfpd(&(u).c13.re)

#define _dhs_su3_load(u,a) \
                    su##a##11 = __lfps(&(u).c11.re);\
                su##a##22 = __lfps(&(u).c22.re);\
                su##a##33 = __lfps(&(u).c33.re);\
                su##a##12 = __lfps(&(u).c12.re);\
                su##a##23 = __lfps(&(u).c23.re);\
                su##a##31 = __lfps(&(u).c31.re);\
                su##a##21 = __lfps(&(u).c21.re);\
                su##a##32 = __lfps(&(u).c32.re);\
                su##a##13 = __lfps(&(u).c13.re)


/*
* SU(3) matrix stored in suc times SU(3) vector ua and ub in parallel
*
* dcra1=(suc*ua).c1
* dcra2=(suc*ua).c2
* dcra3=(suc*ua).c3 and equiv. for dcrb...
*
*/
#define _dh_su3_multiply(a,b,c) \
		t##a##1  = __fxpmul(su##c##11, __creal(u##a##1)); \
		t##a##2  = __fxpmul(su##c##22, __creal(u##a##2)); \
		t##a##3  = __fxpmul(su##c##33, __creal(u##a##3)); \
		\
		t##b##1  = __fxpmul(su##c##12, __creal(u##b##2)); \
		t##b##2  = __fxpmul(su##c##23, __creal(u##b##3)); \
		t##b##3  = __fxpmul(su##c##31, __creal(u##b##1)); \
		\
		dcr##a##1  = __fxcxnpma(t##a##1, su##c##11, __cimag(u##a##1)); \
		dcr##a##2  = __fxcxnpma(t##a##2, su##c##22, __cimag(u##a##2)); \
		dcr##a##3  = __fxcxnpma(t##a##3, su##c##33, __cimag(u##a##3)); \
		\
		dcr##b##1 = __fxcxnpma(t##b##1, su##c##12, __cimag(u##b##2)); \
		dcr##b##2 = __fxcxnpma(t##b##2, su##c##23, __cimag(u##b##3)); \
		dcr##b##3 = __fxcxnpma(t##b##3, su##c##31, __cimag(u##b##1)); \
		\
		t##a##1  = __fxcpmadd(dcr##a##2, su##c##21, __creal(u##a##1)); \
		t##a##2  = __fxcpmadd(dcr##a##3, su##c##32, __creal(u##a##2)); \
		t##a##3  = __fxcpmadd(dcr##a##1, su##c##13, __creal(u##a##3)); \
		\
		t##b##1  = __fxcpmadd(dcr##b##2, su##c##22, __creal(u##b##2)); \
		t##b##2  = __fxcpmadd(dcr##b##3, su##c##33, __creal(u##b##3)); \
		t##b##3  = __fxcpmadd(dcr##b##1, su##c##11, __creal(u##b##1)); \
		\
		dcr##a##2  = __fxcxnpma(t##a##1, su##c##21, __cimag(u##a##1)); \
		dcr##a##3  = __fxcxnpma(t##a##2, su##c##32, __cimag(u##a##2)); \
		dcr##a##1  = __fxcxnpma(t##a##3, su##c##13, __cimag(u##a##3)); \
		\
		dcr##b##2 = __fxcxnpma(t##b##1, su##c##22, __cimag(u##b##2)); \
		dcr##b##3 = __fxcxnpma(t##b##2, su##c##33, __cimag(u##b##3)); \
		dcr##b##1 = __fxcxnpma(t##b##3, su##c##11, __cimag(u##b##1)); \
		\
		t##a##1  = __fxcpmadd(dcr##a##3, su##c##31, __creal(u##a##1)); \
		t##a##2  = __fxcpmadd(dcr##a##1, su##c##12, __creal(u##a##2)); \
		t##a##3  = __fxcpmadd(dcr##a##2, su##c##23, __creal(u##a##3)); \
		\
		t##b##1  = __fxcpmadd(dcr##b##3, su##c##32, __creal(u##b##2)); \
		t##b##2  = __fxcpmadd(dcr##b##1, su##c##13, __creal(u##b##3)); \
		t##b##3  = __fxcpmadd(dcr##b##2, su##c##21, __creal(u##b##1)); \
		\
		dcr##a##3  = __fxcxnpma(t##a##1, su##c##31, __cimag(u##a##1)); \
		dcr##a##1  = __fxcxnpma(t##a##2, su##c##12, __cimag(u##a##2)); \
		dcr##a##2  = __fxcxnpma(t##a##3, su##c##23, __cimag(u##a##3)); \
		\
		dcr##b##3 = __fxcxnpma(t##b##1, su##c##32, __cimag(u##b##2)); \
		dcr##b##1 = __fxcxnpma(t##b##2, su##c##13, __cimag(u##b##3)); \
		dcr##b##2 = __fxcxnpma(t##b##3, su##c##21, __cimag(u##b##1))

/*
* SU(3) matrix stored in suc^dagger times SU(3) vector ua and ub in parallel
*
* dcra1=(suc^dagger*ua).c1
* dcra2=(suc^dagger*ua).c2
* dcra3=(suc^dagger*ua).c3 and equiv. for dcrb...
*
*/
#define _dh_su3_inverse_multiply(a,b,c) \
		t##a##1  = __fxpmul(u##a##1, __creal(su##c##11)); \
		t##a##2  = __fxpmul(u##a##2, __creal(su##c##22)); \
		t##a##3  = __fxpmul(u##a##3, __creal(su##c##33)); \
		\
		t##b##1  = __fxpmul(u##b##2, __creal(su##c##21)); \
		t##b##2  = __fxpmul(u##b##3, __creal(su##c##32)); \
		t##b##3  = __fxpmul(u##b##1, __creal(su##c##13)); \
		\
		dcr##a##1  = __fxcxnsma(t##a##1, u##a##1, __cimag(su##c##11)); \
		dcr##a##2  = __fxcxnsma(t##a##2, u##a##2, __cimag(su##c##22)); \
		dcr##a##3  = __fxcxnsma(t##a##3, u##a##3, __cimag(su##c##33)); \
		\
		dcr##b##1 = __fxcxnsma(t##b##1, u##b##2, __cimag(su##c##21)); \
		dcr##b##2 = __fxcxnsma(t##b##2, u##b##3, __cimag(su##c##32)); \
		dcr##b##3 = __fxcxnsma(t##b##3, u##b##1, __cimag(su##c##13)); \
		\
		t##a##1  = __fxcpmadd(dcr##a##2, u##a##1, __creal(su##c##12)); \
		t##a##2  = __fxcpmadd(dcr##a##3, u##a##2, __creal(su##c##23)); \
		t##a##3  = __fxcpmadd(dcr##a##1, u##a##3, __creal(su##c##31)); \
		\
		t##b##1  = __fxcpmadd(dcr##b##2, u##b##2, __creal(su##c##22)); \
		t##b##2  = __fxcpmadd(dcr##b##3, u##b##3, __creal(su##c##33)); \
		t##b##3  = __fxcpmadd(dcr##b##1, u##b##1, __creal(su##c##11)); \
		\
		dcr##a##2  = __fxcxnsma(t##a##1, u##a##1, __cimag(su##c##12)); \
		dcr##a##3  = __fxcxnsma(t##a##2, u##a##2, __cimag(su##c##23)); \
		dcr##a##1  = __fxcxnsma(t##a##3, u##a##3, __cimag(su##c##31)); \
		\
		dcr##b##2 = __fxcxnsma(t##b##1, u##b##2, __cimag(su##c##22)); \
		dcr##b##3 = __fxcxnsma(t##b##2, u##b##3, __cimag(su##c##33)); \
		dcr##b##1 = __fxcxnsma(t##b##3, u##b##1, __cimag(su##c##11)); \
		\
		t##a##1  = __fxcpmadd(dcr##a##3, u##a##1, __creal(su##c##13)); \
		t##a##2  = __fxcpmadd(dcr##a##1, u##a##2, __creal(su##c##21)); \
		t##a##3  = __fxcpmadd(dcr##a##2, u##a##3, __creal(su##c##32)); \
		\
		t##b##1  = __fxcpmadd(dcr##b##3, u##b##2, __creal(su##c##23)); \
		t##b##2  = __fxcpmadd(dcr##b##1, u##b##3, __creal(su##c##31)); \
		t##b##3  = __fxcpmadd(dcr##b##2, u##b##1, __creal(su##c##12)); \
		\
		dcr##a##3  = __fxcxnsma(t##a##1, u##a##1, __cimag(su##c##13)); \
		dcr##a##1  = __fxcxnsma(t##a##2, u##a##2, __cimag(su##c##21)); \
		dcr##a##2  = __fxcxnsma(t##a##3, u##a##3, __cimag(su##c##32)); \
		\
		dcr##b##3 = __fxcxnsma(t##b##1, u##b##2, __cimag(su##c##23)); \
		dcr##b##1 = __fxcxnsma(t##b##2, u##b##3, __cimag(su##c##31)); \
		dcr##b##2 = __fxcxnsma(t##b##3, u##b##1, __cimag(su##c##12))



/*******************************************************************************
*
* Weyl vector manipulations
*
*******************************************************************************/

/*
	Temporary variables to store 3 Weyl vectors
*/
#define _dh_weyl_vector_declare(a) \
                double _Complex dcr##a##1,dcr##a##2,dcr##a##3,dcr##a##4,dcr##a##5,dcr##a##6;\
                double _Complex t##a##1,t##a##2,t##a##3,t##a##4,t##a##5,t##a##6;\
                double _Complex u##a##1,u##a##2,u##a##3,u##a##4,u##a##5,u##a##6

/*
	Load 1 Weyl vector into ua1..6
*/
#define _dh_weyl_vector_loadu(s,a) \
                u##a##1 = __lfpd(&(s).c1.c1.re); \
                u##a##2 = __lfpd(&(s).c1.c2.re); \
                u##a##3 = __lfpd(&(s).c1.c3.re); \
                u##a##4 = __lfpd(&(s).c2.c1.re); \
                u##a##5 = __lfpd(&(s).c2.c2.re); \
                u##a##6 = __lfpd(&(s).c2.c3.re)

#define _dhs_weyl_vector_loadu(s,a) \
                u##a##1 = __lfps(&(s).c1.c1.re); \
                u##a##2 = __lfps(&(s).c1.c2.re); \
                u##a##3 = __lfps(&(s).c1.c3.re); \
                u##a##4 = __lfps(&(s).c2.c1.re); \
                u##a##5 = __lfps(&(s).c2.c2.re); \
                u##a##6 = __lfps(&(s).c2.c3.re)

/*
	Load 1 Weyl vector into spinor components dcra1..3 and dcrb1..3
*/
#define _dh_weyl_vector_load2(s,a,b) \
                dcr##a##1 = __lfpd(&(s).c1.c1.re); \
                dcr##a##2 = __lfpd(&(s).c1.c2.re); \
                dcr##a##3 = __lfpd(&(s).c1.c3.re); \
                dcr##b##1 = __lfpd(&(s).c2.c1.re); \
                dcr##b##2 = __lfpd(&(s).c2.c2.re); \
                dcr##b##3 = __lfpd(&(s).c2.c3.re)

#define _dhs_weyl_vector_load2(s,a,b) \
                dcr##a##1 = __lfps(&(s).c1.c1.re); \
                dcr##a##2 = __lfps(&(s).c1.c2.re); \
                dcr##a##3 = __lfps(&(s).c1.c3.re); \
                dcr##b##1 = __lfps(&(s).c2.c1.re); \
                dcr##b##2 = __lfps(&(s).c2.c2.re); \
                dcr##b##3 = __lfps(&(s).c2.c3.re)

/*
	ua1..3 = dcrb1..3
	ua4..6 = dcrc1..3
*/
#define _dh_weyl_vector_assign(a,b,c) \
                u##a##1 = dcr##b##1; \
                u##a##2 = dcr##b##2; \
                u##a##3 = dcr##b##3; \
                u##a##4 = dcr##c##1; \
                u##a##5 = dcr##c##2; \
                u##a##6 = dcr##c##3

/*
	dcrb1..3=dcra1..3
	dcrc1..3=dcra4..5
*/
#define _dh_vector_weyl_assign(b,c,a) \
                dcr##c##3 = dcr##a##6; \
                dcr##b##1 = dcr##a##1; \
                dcr##b##2 = dcr##a##2; \
                dcr##b##3 = dcr##a##3; \
                dcr##c##1 = dcr##a##4; \
                dcr##c##2 = dcr##a##5

/*
	Store dcra1..6 into Weyl vector s
*/
#define _dh_weyl_vector_store(s,a) \
                __stfpd(&(s).c2.c3.re,  dcr##a##6);\
                __stfpd(&(s).c1.c1.re,  dcr##a##1); \
                __stfpd(&(s).c1.c2.re,  dcr##a##2); \
                __stfpd(&(s).c1.c3.re,  dcr##a##3); \
                __stfpd(&(s).c2.c1.re,  dcr##a##4); \
                __stfpd(&(s).c2.c2.re,  dcr##a##5)

#define _dhs_weyl_vector_store(s,a) \
                __stfps(&(s).c2.c3.re,  dcr##a##6);\
                __stfps(&(s).c1.c1.re,  dcr##a##1); \
                __stfps(&(s).c1.c2.re,  dcr##a##2); \
                __stfps(&(s).c1.c3.re,  dcr##a##3); \
                __stfps(&(s).c2.c1.re,  dcr##a##4); \
                __stfps(&(s).c2.c2.re,  dcr##a##5)

/*
	Store dcra1..3 and dcrb1..3 into Weyl vector s
*/
#define _dh_weyl_vector_store2(s,a,b) \
                __stfpd(&(s).c1.c1.re,  dcr##a##1); \
                __stfpd(&(s).c1.c2.re,  dcr##a##2); \
                __stfpd(&(s).c1.c3.re,  dcr##a##3); \
                __stfpd(&(s).c2.c1.re,  dcr##b##1); \
                __stfpd(&(s).c2.c2.re,  dcr##b##2); \
                __stfpd(&(s).c2.c3.re,  dcr##b##3)

#define _dhs_weyl_vector_store2(s,a,b) \
                __stfps(&(s).c1.c1.re,  dcr##a##1); \
                __stfps(&(s).c1.c2.re,  dcr##a##2); \
                __stfps(&(s).c1.c3.re,  dcr##a##3); \
                __stfps(&(s).c2.c1.re,  dcr##b##1); \
                __stfps(&(s).c2.c2.re,  dcr##b##2); \
                __stfps(&(s).c2.c3.re,  dcr##b##3)


/*******************************************************************************
*
* Pauli matrix manipulations
*
*******************************************************************************/

/*
	Temporary variables to store a (hermitian) Pauli matrix
*/
#define _dh_pauli_matrix_declare(a) \
             double _Complex m##a##1122, m##a##3344,  m##a##5566,\
		    m##a##12, m##a##13, m##a##14, m##a##15, m##a##16,\
		              m##a##23, m##a##24, m##a##25, m##a##26,\
		                        m##a##34, m##a##35, m##a##36,\
				                  m##a##45, m##a##46,\
				                            m##a##56

/*
	Load pauli matrix into temporary variables
*/
#define _dh_pauli_load(p,a) \
                m##a##1122 = __lfpd(&(p).u[0]);\
                m##a##3344 = __lfpd(&(p).u[2]);\
                m##a##5566 = __lfpd(&(p).u[4]);\
                m##a##12 = __lfpd(&(p).u[6]);\
                m##a##13 = __lfpd(&(p).u[8]);\
                m##a##14 = __lfpd(&(p).u[10]);\
                m##a##15 = __lfpd(&(p).u[12]);\
                m##a##16 = __lfpd(&(p).u[14]);\
                m##a##23 = __lfpd(&(p).u[16]);\
                m##a##24 = __lfpd(&(p).u[18]);\
                m##a##25 = __lfpd(&(p).u[20]);\
                m##a##26 = __lfpd(&(p).u[22]);\
                m##a##34 = __lfpd(&(p).u[24]);\
                m##a##35 = __lfpd(&(p).u[26]);\
                m##a##36 = __lfpd(&(p).u[28]);\
                m##a##45 = __lfpd(&(p).u[30]);\
                m##a##46 = __lfpd(&(p).u[32]);\
                m##a##56 = __lfpd(&(p).u[34])

#define _dhs_pauli_load(p,a) \
                m##a##1122 = __lfps(&(p).u[0]);\
                m##a##3344 = __lfps(&(p).u[2]);\
                m##a##5566 = __lfps(&(p).u[4]);\
                m##a##12 = __lfps(&(p).u[6]);\
                m##a##13 = __lfps(&(p).u[8]);\
                m##a##14 = __lfps(&(p).u[10]);\
                m##a##15 = __lfps(&(p).u[12]);\
                m##a##16 = __lfps(&(p).u[14]);\
                m##a##23 = __lfps(&(p).u[16]);\
                m##a##24 = __lfps(&(p).u[18]);\
                m##a##25 = __lfps(&(p).u[20]);\
                m##a##26 = __lfps(&(p).u[22]);\
                m##a##34 = __lfps(&(p).u[24]);\
                m##a##35 = __lfps(&(p).u[26]);\
                m##a##36 = __lfps(&(p).u[28]);\
                m##a##45 = __lfps(&(p).u[30]);\
                m##a##46 = __lfps(&(p).u[32]);\
                m##a##56 = __lfps(&(p).u[34])


/*
* Pauli matrix stored in mc times Weyl vector dcra
*
* u1=(mc*dcra).c1
* u2=(mc*dcra).c2
* u3=(mc*dcra).c3
* u4=(mc*dcra).c4
* u5=(mc*dcra).c5
* u6=(mc*dcra).c6
*
*/
#define _dh_mul_pauli(a,c) \
		dcr##a##1  = __fxpmul(u##a##1, __creal(m##c##1122)); \
		dcr##a##2  = __fxpmul(u##a##2, __cimag(m##c##1122)); \
		dcr##a##3  = __fxpmul(u##a##3, __creal(m##c##3344)); \
		dcr##a##4  = __fxpmul(u##a##4, __cimag(m##c##3344)); \
		dcr##a##5  = __fxpmul(u##a##5, __creal(m##c##5566)); \
		dcr##a##6  = __fxpmul(u##a##6, __cimag(m##c##5566)); \
		\
		t##a##1  = __fxcpmadd(dcr##a##2, u##a##1, __creal(m##c##12)); \
		t##a##2  = __fxcpmadd(dcr##a##3, u##a##2, __creal(m##c##23)); \
		t##a##3  = __fxcpmadd(dcr##a##4, u##a##3, __creal(m##c##34)); \
		t##a##4  = __fxcpmadd(dcr##a##5, u##a##4, __creal(m##c##45)); \
		t##a##5  = __fxcpmadd(dcr##a##6, u##a##5, __creal(m##c##56)); \
		t##a##6  = __fxcpmadd(dcr##a##1, u##a##6, __creal(m##c##16)); \
		\
		dcr##a##2  = __fxcxnsma(t##a##1, u##a##1, __cimag(m##c##12)); \
		dcr##a##3  = __fxcxnsma(t##a##2, u##a##2, __cimag(m##c##23)); \
		dcr##a##4  = __fxcxnsma(t##a##3, u##a##3, __cimag(m##c##34)); \
		dcr##a##5  = __fxcxnsma(t##a##4, u##a##4, __cimag(m##c##45)); \
		dcr##a##6  = __fxcxnsma(t##a##5, u##a##5, __cimag(m##c##56)); \
		dcr##a##1  = __fxcxnpma(t##a##6, u##a##6, __cimag(m##c##16)); \
		\
		t##a##1  = __fxcpmadd(dcr##a##3, u##a##1, __creal(m##c##13)); \
		t##a##2  = __fxcpmadd(dcr##a##4, u##a##2, __creal(m##c##24)); \
		t##a##3  = __fxcpmadd(dcr##a##5, u##a##3, __creal(m##c##35)); \
		t##a##4  = __fxcpmadd(dcr##a##6, u##a##4, __creal(m##c##46)); \
		t##a##5  = __fxcpmadd(dcr##a##1, u##a##5, __creal(m##c##15)); \
		t##a##6  = __fxcpmadd(dcr##a##2, u##a##6, __creal(m##c##26)); \
		\
		dcr##a##3  = __fxcxnsma(t##a##1, u##a##1, __cimag(m##c##13)); \
		dcr##a##4  = __fxcxnsma(t##a##2, u##a##2, __cimag(m##c##24)); \
		dcr##a##5  = __fxcxnsma(t##a##3, u##a##3, __cimag(m##c##35)); \
		dcr##a##6  = __fxcxnsma(t##a##4, u##a##4, __cimag(m##c##46)); \
		dcr##a##1  = __fxcxnpma(t##a##5, u##a##5, __cimag(m##c##15)); \
		dcr##a##2  = __fxcxnpma(t##a##6, u##a##6, __cimag(m##c##26)); \
		\
		t##a##1  = __fxcpmadd(dcr##a##4, u##a##1, __creal(m##c##14)); \
		t##a##2  = __fxcpmadd(dcr##a##5, u##a##2, __creal(m##c##25)); \
		t##a##3  = __fxcpmadd(dcr##a##6, u##a##3, __creal(m##c##36)); \
		t##a##4  = __fxcpmadd(dcr##a##1, u##a##4, __creal(m##c##14)); \
		t##a##5  = __fxcpmadd(dcr##a##2, u##a##5, __creal(m##c##25)); \
		t##a##6  = __fxcpmadd(dcr##a##3, u##a##6, __creal(m##c##36)); \
		\
		dcr##a##4  = __fxcxnsma(t##a##1, u##a##1, __cimag(m##c##14)); \
		dcr##a##5  = __fxcxnsma(t##a##2, u##a##2, __cimag(m##c##25)); \
		dcr##a##6  = __fxcxnsma(t##a##3, u##a##3, __cimag(m##c##36)); \
		dcr##a##1  = __fxcxnpma(t##a##4, u##a##4, __cimag(m##c##14)); \
		dcr##a##2  = __fxcxnpma(t##a##5, u##a##5, __cimag(m##c##25)); \
		dcr##a##3  = __fxcxnpma(t##a##6, u##a##6, __cimag(m##c##36)); \
		\
		t##a##1  = __fxcpmadd(dcr##a##5, u##a##1, __creal(m##c##15)); \
		t##a##2  = __fxcpmadd(dcr##a##6, u##a##2, __creal(m##c##26)); \
		t##a##3  = __fxcpmadd(dcr##a##1, u##a##3, __creal(m##c##13)); \
		t##a##4  = __fxcpmadd(dcr##a##2, u##a##4, __creal(m##c##24)); \
		t##a##5  = __fxcpmadd(dcr##a##3, u##a##5, __creal(m##c##35)); \
		t##a##6  = __fxcpmadd(dcr##a##4, u##a##6, __creal(m##c##46)); \
		\
		dcr##a##5  = __fxcxnsma(t##a##1, u##a##1, __cimag(m##c##15)); \
		dcr##a##6  = __fxcxnsma(t##a##2, u##a##2, __cimag(m##c##26)); \
		dcr##a##1  = __fxcxnpma(t##a##3, u##a##3, __cimag(m##c##13)); \
		dcr##a##2  = __fxcxnpma(t##a##4, u##a##4, __cimag(m##c##24)); \
		dcr##a##3  = __fxcxnpma(t##a##5, u##a##5, __cimag(m##c##35)); \
		dcr##a##4  = __fxcxnpma(t##a##6, u##a##6, __cimag(m##c##46)); \
		\
		t##a##1  = __fxcpmadd(dcr##a##6, u##a##1, __creal(m##c##16)); \
		t##a##2  = __fxcpmadd(dcr##a##1, u##a##2, __creal(m##c##12)); \
		t##a##3  = __fxcpmadd(dcr##a##2, u##a##3, __creal(m##c##23)); \
		t##a##4  = __fxcpmadd(dcr##a##3, u##a##4, __creal(m##c##34)); \
		t##a##5  = __fxcpmadd(dcr##a##4, u##a##5, __creal(m##c##45)); \
		t##a##6  = __fxcpmadd(dcr##a##5, u##a##6, __creal(m##c##56)); \
		\
		dcr##a##6  = __fxcxnsma(t##a##1, u##a##1, __cimag(m##c##16)); \
		dcr##a##1  = __fxcxnpma(t##a##2, u##a##2, __cimag(m##c##12)); \
		dcr##a##2  = __fxcxnpma(t##a##3, u##a##3, __cimag(m##c##23)); \
		dcr##a##3  = __fxcxnpma(t##a##4, u##a##4, __cimag(m##c##34)); \
		dcr##a##4  = __fxcxnpma(t##a##5, u##a##5, __cimag(m##c##45)); \
		dcr##a##5  = __fxcxnpma(t##a##6, u##a##6, __cimag(m##c##56))

#endif

